import { Ref, WatchSource, ShallowUnwrapRef, WatchOptions } from 'vue-demi';
import { ComputedRef, WatchStopHandle } from 'vue-demi';

declare function biSyncRef<R extends Ref<any>>(a: R, b: R): () => void;

/**
 * Explicitly define the deps of computed
 *
 * @param source
 * @param fn
 */
declare function controlledComputed<T, S>(source: WatchSource<S>, fn: () => T): ComputedRef<T>;

interface ExtendRefOptions<Unwrap extends boolean = boolean> {
    /**
     * Is the extends properties enumerable
     *
     * @default false
     */
    enumerable?: boolean;
    /**
     * Unwrap for Ref properties
     *
     * @default true
     */
    unwrap?: Unwrap;
}
/**
 * Overlad 1: Unwrap set to false
 */
declare function extendRef<R extends Ref<any>, Extend extends object, Options extends ExtendRefOptions<false>>(ref: R, extend: Extend, options: Options): ShallowUnwrapRef<Extend> & R;
/**
 * Overlad 2: Unwrap unset or set to true
 */
declare function extendRef<R extends Ref<any>, Extend extends object, Options extends ExtendRefOptions>(ref: R, extend: Extend, options: Options): Extend & R;

declare function makeDestructurable<T extends Record<string, unknown>, A extends readonly any[]>(obj: T, arr: A): T & A;

interface SyncRefOptions {
    /**
     * Timing for syncing, same as watch's flush option
     *
     * @default 'post'
     */
    flush?: WatchOptions['flush'];
    /**
     * Watch deeply
     *
     * @default false
     */
    deep?: boolean;
    /**
     * Sync values immediately
     *
     * @default true
     */
    immediate?: boolean;
}
/**
 * Keep target ref(s) in sync with the source ref
 *
 * @param source source ref
 * @param targets
 */
declare function syncRef<R extends Ref<any>>(source: R, targets: R | R[], { flush, deep, immediate, }?: SyncRefOptions): WatchStopHandle;

/**
 * Call onMounted() if it's inside a component lifecycle, if not, run just call the function
 *
 * @param fn
 * @param sync if set to false, it will run in the nextTick() of Vue
 */
declare function tryOnMounted(fn: () => void, sync?: boolean): void;

/**
 * Call onUnmounted() if it's inside a component lifecycle, if not, do nothing
 *
 * @param fn
 */
declare function tryOnUnmounted(fn: () => void): void;

declare const isClient: boolean;
declare const isDef: <T = any>(val?: T | undefined) => val is T;
declare const assert: (condition: boolean, ...infos: any[]) => void;
declare const isBoolean: (val: any) => val is boolean;
declare const isFunction: <T = Function>(val: any) => val is T;
declare const isNumber: (val: any) => val is number;
declare const isString: (val: unknown) => val is string;
declare const isObject: (val: any) => val is object;
declare const isWindow: (val: any) => val is Window;
declare const now: () => number;
declare const timestamp: () => number;
declare const clamp: (n: number, min: number, max: number) => number;
declare const noop: () => void;

declare type MaybeRef<T> = T | Ref<T>;
declare function promiseTimeout(ms: number, throwOnTimeout?: boolean, reason?: string): Promise<void>;
declare function invoke<T>(fn: () => T): T;

interface WhenToMatchOptions {
    flush?: WatchOptions['flush'];
    timeout?: number;
    throwOnTimeout?: boolean;
}
interface WhenInstance<T> {
    readonly not: WhenInstance<T>;
    toMatch(condition: (v: T | object) => boolean, options?: WhenToMatchOptions): Promise<void>;
    toBe<P>(value: P | T, options?: WhenToMatchOptions): Promise<void>;
    toBeTruthy(options?: WhenToMatchOptions): Promise<void>;
    toBeNull(options?: WhenToMatchOptions): Promise<void>;
    toBeUndefined(options?: WhenToMatchOptions): Promise<void>;
    toBeNaN(options?: WhenToMatchOptions): Promise<void>;
    toContain<P>(value: P, options?: WhenToMatchOptions): Promise<void>;
    changed(options?: WhenToMatchOptions): Promise<void>;
    changedTimes(n?: number, options?: WhenToMatchOptions): Promise<void>;
}
declare function when<T>(r: WatchSource<T> | object): WhenInstance<T>;

export { ExtendRefOptions, MaybeRef, SyncRefOptions, WhenInstance, WhenToMatchOptions, assert, biSyncRef, clamp, controlledComputed, extendRef, invoke, isBoolean, isClient, isDef, isFunction, isNumber, isObject, isString, isWindow, makeDestructurable, noop, now, promiseTimeout, syncRef, timestamp, tryOnMounted, tryOnUnmounted, when };
